/* Copyright (C) 2011, 2012, 2013 by Bundesanstalt für Gewässerkunde
 * Software engineering by Intevation GmbH
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.artifacts.sinfo.flood_duration;

import java.io.OutputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;
import org.dive4elements.river.artifacts.common.GeneralResultType;
import org.dive4elements.river.artifacts.common.JasperDesigner;
import org.dive4elements.river.artifacts.common.JasperReporter;
import org.dive4elements.river.artifacts.common.MetaAndTableJRDataSource;
import org.dive4elements.river.artifacts.common.ResultRow;
import org.dive4elements.river.artifacts.sinfo.common.AbstractSInfoExporter;
import org.dive4elements.river.artifacts.sinfo.common.SInfoResultType;
import org.dive4elements.river.artifacts.sinfo.util.RiverInfo;
import org.dive4elements.river.artifacts.sinfo.util.WstInfo;

import au.com.bytecode.opencsv.CSVWriter;
import net.sf.jasperreports.engine.JRException;

/**
 * Generates different output formats (csv, pdf) of data that resulted from a flow depths min/max computation.
 *
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 * @author Gernot Belger
 */
// REMARK: must be public because its registered in generators.xml
public class FloodDurationExporter extends AbstractSInfoExporter<FloodDurationCalculationResult, FloodDurationCalculationResults> {

    /** The log used in this exporter. */
    private static Logger log = Logger.getLogger(FloodDurationExporter.class);

    private final String getJasperFile(final int waterlevelCount) {
        if (waterlevelCount <= 1)
            return "/jasper/templates/sinfo.floodduration.jrxml"; // TODO use jrxml-path all over the project
        else
            return "/jasper/templates/sinfo.floodduration2.jrxml";
    }

    private static final int maxPdfWspls = 3;

    @Override
    protected Logger getLog() {
        return log;
    }

    @Override
    protected void writeCSVGlobalMetadata(final CSVWriter writer, final FloodDurationCalculationResults results) {
        log.info("FloodDurationExporter.writeCSVMeta");
        super.writeCSVGlobalMetadataDefaults(writer, results);
    }

    @Override
    protected void writeCSVResultMetadata(final CSVWriter writer, final FloodDurationCalculationResults results, final FloodDurationCalculationResult result) {

        final WstInfo wst = result.getWst();
        super.writeCSVWaterlevelMetadata(writer, wst); // TODO: Abweichend vom Allgemeinen werden hier andere Felder benötigt bei den Wasserspiegellagen

        writer.writeNext(new String[] { "" }); // break line

    }

    /**
     * Write the header, with different headings depending on whether at a
     * gauge or at a location.
     *
     * @param river
     * @param useTkh
     */
    @Override
    protected void writeCSVHeader(final CSVWriter writer, final FloodDurationCalculationResults results, final RiverInfo river) {
        log.info("FloodDurationExporter.writeCSVHeader");

        final Collection<String> header = new ArrayList<>(99);

        header.add(msg(GeneralResultType.station.getCsvHeader()));
        header.add(msg(SInfoResultType.riverside.getCsvHeader()));
        header.add(msg(SInfoResultType.inundationduration.getCsvHeader()));
        header.add(msg(SInfoResultType.inundationdurationq.getCsvHeader()));
        header.add(msg(SInfoResultType.infrastructureHeight.getCsvHeader()));
        header.add(msg(SInfoResultType.infrastructuretype.getCsvHeader()));
        // add dynamic headers
        final int waterlevelCount = getWaterlevelCount();
        for (int i = 0; i < waterlevelCount; i++) {
            final int naturalIndex = i + 1;
            final String appendIndex = new StringBuilder().append("_").append(naturalIndex).toString();
            final Object[] args = new Object[] { appendIndex };
            // new StringBuilder().append('\u2081').toString(); // schlechter UTF-8-Support für subscript ints
            header.add(msg(DurationWaterlevel.getHeaderW(), new Object[] { appendIndex, "results.getRiver().getWstUnit()" }));
            header.add(msg(DurationWaterlevel.getHeaderFloodDurPerYear(), args));
            header.add(msg(DurationWaterlevel.getHeaderQ(), args));
            header.add(msg(DurationWaterlevel.getHeaderBezeichn(), args));
        }

        header.add(msg(SInfoResultType.gaugeLabel.getCsvHeader()));
        header.add(msg(SInfoResultType.location.getCsvHeader()));

        writer.writeNext(header.toArray(new String[header.size()]));
    }

    private int getWaterlevelCount() {
        final FloodDurationCalculationResults results = getData();
        if (results != null) {
            final List<FloodDurationCalculationResult> list = results.getResults();
            if (list != null && list.size() > 0) {
                final FloodDurationCalculationResult result = list.get(0);
                return result.getWaterlevelCount();
            }
        }
        return 0;
    }

    @Override
    protected String[] formatRow(final FloodDurationCalculationResults results, final ResultRow row, final ExportMode mode) {

        final Collection<String> lines = new ArrayList<>(99);

        lines.add(row.exportValue(this.context, GeneralResultType.station));
        lines.add(row.exportValue(this.context, SInfoResultType.riverside));
        lines.add(row.exportValue(this.context, SInfoResultType.inundationduration));
        lines.add(row.exportValue(this.context, SInfoResultType.inundationdurationq));
        lines.add(row.exportValue(this.context, SInfoResultType.infrastructureHeight));
        lines.add(row.exportValue(this.context, SInfoResultType.infrastructuretype));

        final int waterlevelcount = this.getWaterlevelCount();

        final List<DurationWaterlevel> waterlevelList = (List<DurationWaterlevel>) row.getValue(SInfoResultType.customMultiRowColWaterlevel);
        if (waterlevelList != null) {
            final NumberFormat wFormatter = getFlowDepthFormatter();
            final NumberFormat qFormatter = getQFormatter();

            for (int i = 0; i < waterlevelList.size(); i++) {

                if (i == FloodDurationExporter.maxPdfWspls && mode == ExportMode.pdf)
                    break;

                final DurationWaterlevel item = waterlevelList.get(i);
                lines.add(item.getWFormatted(wFormatter));
                lines.add(item.getFloodDurDaysPerYearFormatted());
                lines.add(item.getQFormatted(qFormatter));
                lines.add(item.getBezeichnung());
            }
        }

        if ((waterlevelcount == 0 || waterlevelcount == 2) && mode == ExportMode.pdf) {
            lines.add("dummy");
            lines.add("dummy");
            lines.add("dummy");
            lines.add("dummy");
        }

        lines.add(row.exportValue(this.context, SInfoResultType.gaugeLabel));
        lines.add(row.exportValue(this.context, SInfoResultType.location));
        return lines.toArray(new String[lines.size()]);
    }

    @Override
    protected void writePDF(final OutputStream out) {

        try {
            final MetaAndTableJRDataSource source = createJRData(this.data);
            final JasperReporter reporter = new JasperReporter();
            final int waterlevelCount = getWaterlevelCount();
            final JasperDesigner design = reporter.addReport(getJasperFile(waterlevelCount), source);

            if (waterlevelCount == 0 || waterlevelCount == 2) {
                design.removeColumn("wOpt");
                design.removeColumn("qOpt");
                design.removeColumn("bezOpt");
                design.removeColumn("durOpt");
            }

            reporter.exportPDF(out);
        }
        catch (final JRException je) {
            getLog().warn("Error generating PDF Report!", je);
        }
    }

    @Override
    protected final void addJRMetaData(final MetaAndTableJRDataSource source, final FloodDurationCalculationResults results) {

        /* general metadata */
        super.addJRMetaData(source, results);

        /* column headings */
        source.addMetaData("station_header", GeneralResultType.station.getPdfHeader(this.context.getMeta()));
        source.addMetaData("riverside_header", SInfoResultType.riverside.getPdfHeader(this.context.getMeta()));
        source.addMetaData("inundationduration_header", SInfoResultType.inundationduration.getPdfHeader(this.context.getMeta()));
        source.addMetaData("inundationduration_q_header", SInfoResultType.inundationdurationq.getPdfHeader(this.context.getMeta()));
        source.addMetaData("infrastructure_height_header", SInfoResultType.infrastructureHeightFloodDur.getPdfHeader(this.context.getMeta()));
        source.addMetaData("infrastructure_type_header", SInfoResultType.infrastructuretype.getPdfHeader(this.context.getMeta()));

        // add dynamic headers

        final int waterlevelCount = getWaterlevelCount() > FloodDurationExporter.maxPdfWspls ? FloodDurationExporter.maxPdfWspls : getWaterlevelCount();

        if (waterlevelCount == 0 || waterlevelCount == 2) {
            source.addMetaData("dummy", "dummy");
            source.addMetaData("dummy", "dummy");
            source.addMetaData("dummy", "dummy");
            source.addMetaData("dummy", "dummy");
        }

        for (int i = 0; i < waterlevelCount; i++) {
            final int naturalIndex = i + 1;

            final Object[] args = new String[] { new StringBuilder().append("_").append(naturalIndex).toString() };
            source.addMetaData(getPdfHeader("w", naturalIndex), msg(DurationWaterlevel.getHeaderW(), args));
            source.addMetaData(getPdfHeader("duration", naturalIndex), msg(DurationWaterlevel.getHeaderFloodDurPerYear(), args));
            source.addMetaData(getPdfHeader("q", naturalIndex), msg(DurationWaterlevel.getHeaderQ(), args));
            source.addMetaData(getPdfHeader("bezeichnung", naturalIndex), msg(DurationWaterlevel.getHeaderBezeichn(), args));

        }

        source.addMetaData("gauge_header", SInfoResultType.gaugeLabel.getPdfHeader(this.context.getMeta()));
        source.addMetaData("location_header", SInfoResultType.location.getPdfHeader(this.context.getMeta()));

    }

    private String getPdfHeader(final String rootStr, final int index) {
        final String hd = "_header";
        final StringBuilder builder = new StringBuilder();
        return builder.append(rootStr).append("_").append(index).append(hd).toString();
    }

}